pub mod dm_type;
pub mod dvm_value;
mod mem;
pub mod module;
mod object_type;
pub mod op_codes;
pub mod platform;
pub mod util;

use crate::vm::dvm_value::DvmValue;
use crate::vm::module::DmModule;
use crate::vm::object_type::DmFn;
use crate::vm::platform::init_platform_functions;
use op_codes::*;
use std::collections::HashMap;
use util::{get_32_le, get_64_le};

pub type PlatformFunction = fn(args: Vec<DvmValue>, &mut DmVirtualMachine) -> DvmValue;

enum CallFrame {
    PlatformCall(PlatformCallFrame),
    DeimosCall(DeimosCallFrame),
}

pub struct PlatformCallFrame {
    pub name: String,
    pub args: Vec<DvmValue>,
}

struct DeimosCallFrame {
    return_address: usize,
}

pub struct DmVirtualMachine {
    modules: Vec<DmModule>,
    platform_functions: HashMap<String, PlatformFunction>,
    ip: usize,
    call_stack: Vec<CallFrame>,
    registers: Vec<DvmValue>,
    register_state_stack: Vec<Vec<DvmValue>>,
}

impl DmVirtualMachine {
    pub fn new(modules: Vec<DmModule>) -> DmVirtualMachine {
        DmVirtualMachine {
            modules,
            ip: 0,
            registers: Vec::new(),
            platform_functions: init_platform_functions(),
            call_stack: Vec::new(),
            register_state_stack: Vec::new(),
        }
    }

    pub fn call_by_fqn(&mut self, fn_fqn: &str, args: Vec<DvmValue>) -> DvmValue {
        todo!()
    }

    pub fn call_fn(&mut self, dm_function: &DmFn, args: Vec<DvmValue>) -> DvmValue {
        // save current state
        self.call_stack.push(CallFrame::DeimosCall(DeimosCallFrame {
            return_address: self.ip,
        }));
        self.register_state_stack.push(self.registers.clone());

        // zero registers and make sure there are enough for dm_function
        self.registers.clear();
        self.registers.resize(args.len(), DvmValue::Unit);

        // push args
        for i in 0..args.len() {
            self.registers.insert(i, args[i].clone());
        }

        // run the byte code
        self.run_raw(&dm_function.byte_code);

        // restore state
        self.registers = self.register_state_stack.pop().unwrap();
        if let CallFrame::DeimosCall(deimos_call_frame) = self.call_stack.pop().unwrap() {
            self.ip = deimos_call_frame.return_address;
        }

        // return result
        self.registers.get(0).unwrap().clone()
    }

    pub fn run_raw(&mut self, code: &Vec<u8>) {
        let mut i = 0;
        while i < code.len() {
            match code[i] {
                MOV_INT => {
                    let target_register = code[i + 1] as usize;
                    let operand = get_32_le!(code, i, 2, u32);
                    self.registers
                        .insert(target_register, DvmValue::Int(operand as i32));
                    i += 6;
                }
                MOV_LONG => {
                    let target_register = code[i + 1] as usize;
                    let operand = get_64_le!(code, i, 2, u64);
                    self.registers
                        .insert(target_register, DvmValue::Long(operand as i64));
                    i += 10;
                }
                MOV_DOUBLE => { /* todo */ }
                MOV_REGISTER => {
                    let target_register = code[i + 1] as usize;
                    let source_register = code[i + 2] as usize;
                    let source_value = self.registers.get(source_register).unwrap();
                    self.registers.insert(target_register, source_value.clone());
                    i += 3;
                }
                ALLOC => {
                    todo!()
                }
                DEALLOC => {
                    todo!()
                }
                MOV_INT_TO => {
                    todo!()
                }
                PLATFORM_CALL => {
                    i += 1;

                    let symbol_name_length = get_32_le!(code, i, 0, usize);
                    i += 4;

                    let symbol_name_raw = code[i..(i + symbol_name_length)].to_vec();
                    i += symbol_name_length;

                    let symbol_name = String::from_utf8(symbol_name_raw).unwrap();

                    let return_register = code[i + 1] as usize;
                    i += 1;

                    let arg_registers_length = code[i] as usize;
                    i += 1;

                    let arg_registers = code[i..(i + arg_registers_length)].to_vec();
                    i += arg_registers_length;

                    let mut args = Vec::new();
                    for arg_register in arg_registers {
                        let register_value = self.registers.get(arg_register as usize).unwrap();
                        args.push(register_value.clone());
                    }

                    let platform_function_result = self.platform_functions.get(&symbol_name);
                    if platform_function_result.is_none() {
                        panic!("Unknown platform function {}", symbol_name);
                    }
                    let platform_function = platform_function_result.unwrap();

                    let call_result = platform_function(args, self);
                    self.registers.insert(return_register, call_result);
                }
                _ => panic!("Invalid code instruction"),
            }
        }
    }
}

#[cfg(test)]
mod dvm_run_tests {
    use super::*;

    macro_rules! assert_register {
        ( $expected: expr, $register: expr ) => {
            assert_eq!($expected, $register.unwrap().clone());
        };
    }

    #[test]
    fn mov_1_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 1);
        let mut vm = DmVirtualMachine::new(Vec::new());
        vm.run_raw(&code);
        assert_register!(DvmValue::Int(1), vm.registers.get(0));
    }

    #[test]
    fn move_65535_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 0xffff);
        let mut vm = DmVirtualMachine::new(Vec::new());
        vm.run_raw(&code);
        assert_register!(DvmValue::Int(0xffff), vm.registers.get(0));
    }

    #[test]
    fn move_int_max_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 0x0fff_ffff);
        let mut vm = DmVirtualMachine::new(Vec::new());
        vm.run_raw(&code);
        assert_register!(DvmValue::Int(0x0fff_ffff), vm.registers.get(0));
    }

    #[test]
    fn move_register() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 1, 1);
        add_mov_register(&mut code, 0, 1);
        let mut vm = DmVirtualMachine::new(Vec::new());
        vm.registers.resize(2, DvmValue::Unit);
        vm.run_raw(&code);
        assert_register!(DvmValue::Int(1), vm.registers.get(0));
    }

    #[test]
    fn mov_int_to_register_as_address() {
        let mut code = Vec::new();
        add_alloc(&mut code, 0, 4);
        add_mov_int_to(&mut code, 0, 0, 0xff);
        let mut vm = DmVirtualMachine::new(Vec::new());
        vm.run_raw(&code);
        let actual = vm.registers.get(0).unwrap().clone();
        match actual {
            DvmValue::Pointer(value_box) => {
                assert_eq!(DvmValue::Int(0xff), *value_box);
            }
            _ => panic!("Target register is not a pointer"),
        }
    }

    #[test]
    fn alloc_and_dealloc_expect_register_cleared() {
        let mut code = Vec::new();
        add_alloc(&mut code, 0, 4);
        add_dealloc(&mut code, 0);
        let mut vm = DmVirtualMachine::new(Vec::new());
        vm.run_raw(&code);
        assert_register!(DvmValue::Unit, vm.registers.get(0));
    }
}
