use crate::vm::dm_type::DmPrimitiveType;
use crate::vm::mem::DmAllocObject;
use std::collections::HashMap;
use std::fmt::Debug;
use std::rc::Rc;

#[derive(Debug, PartialEq, Eq)]
pub enum DmObjectType {
    Interface(DmInterface),
    Implementation(DmImplementation),
}

#[derive(Debug, Eq)]
pub struct DmFn {
    pub fqn: String,
    pub short_name: String,
    pub byte_code: Vec<u8>,
    pub num_registers: usize,
}

impl PartialEq for DmFn {
    fn eq(&self, other: &Self) -> bool {
        self.fqn == other.fqn
    }
}

#[derive(Debug, Eq)]
pub struct DmInterface {
    fqn: String,
    short_name: String,
    properties: HashMap<String, DmProperty>,
    virtual_methods: HashMap<String, DmVirtualMethod>,
    impl_methods: HashMap<String, DmFn>,
}

impl PartialEq for DmInterface {
    fn eq(&self, other: &Self) -> bool {
        self.fqn == other.fqn
    }
}

impl DmInterface {
    fn get_fqn(&self) -> &str {
        &self.fqn
    }

    fn get_method<'a>(&'a self, name: &str, self_object: &'a DmAllocObject) -> Option<&'a DmFn> {
        if self.impl_methods.contains_key(name) {
            self.impl_methods.get(name)
        } else if self.virtual_methods.contains_key(name) {
            self_object.implementation.get_method(name, self_object)
        } else {
            None
        }
    }

    fn get_property(&self, name: &str, self_object: &DmAllocObject) -> Option<&DmProperty> {
        self.properties.get(name)
    }
}

#[derive(Debug, Eq)]
pub struct DmVirtualMethod {
    pub fqn: String,
    pub short_name: String,
}

impl PartialEq for DmVirtualMethod {
    fn eq(&self, other: &Self) -> bool {
        self.fqn == other.fqn
    }
}

#[derive(Debug, Eq)]
pub struct DmImplementation {
    pub fqn: String,
    pub short_name: String,
    pub interface: Rc<DmInterface>,
    pub properties: HashMap<String, DmProperty>,
    pub fields: HashMap<String, DmField>,
    pub size_in_bytes: usize,
    pub methods: HashMap<String, DmFn>,
}

impl PartialEq for DmImplementation {
    fn eq(&self, other: &Self) -> bool {
        self.fqn == other.fqn
    }
}

impl DmImplementation {
    fn get_fqn(&self) -> &str {
        &self.fqn
    }

    fn get_method<'a>(&'a self, name: &str, self_object: &'a DmAllocObject) -> Option<&'a DmFn> {
        self.methods.get(name)
    }

    fn get_property(&self, name: &str, self_object: &DmAllocObject) -> Option<&DmProperty> {
        self.properties.get(name)
    }
}

#[derive(Debug, Eq)]
pub struct DmProperty {
    pub name: String,
    pub data_offset: usize,
    pub primitive_type: Rc<DmPrimitiveType>,
}

impl PartialEq for DmProperty {
    fn eq(&self, other: &Self) -> bool {
        self.name == other.name
    }
}

#[derive(Debug, Eq)]
pub struct DmField {
    name: String,
    primitive_type: Rc<DmPrimitiveType>,
}

impl PartialEq for DmField {
    fn eq(&self, other: &Self) -> bool {
        self.name == other.name
    }
}
